///////////////////////////////////////////////////////////////
//  queue.cpp   --  test multi-threaded queuing              //
//  ver 1.0                                                  //
//  Language:       Visual C++, ver 6.0                      //
//  Platform:       Micron Dual Pentium Pro 200, Win NT 4.0  //
//  Application:    Demo for CSE691, Pr#3, Fall 98           //
//  Author:         Jim Fawcett, CST 2-187, Syracuse Univ.   //
//                  (315) 443-3948, fawcett@ecs.syr.edu      //
///////////////////////////////////////////////////////////////
//  Operations:                                              //
///////////////////////////////////////////////////////////////
//  This program demonstrates synchronizing access to a      //
//  queue by an enQing thread and a deQing thread.  It also  //
//  demonstrates synchronizing access to an output stream    //
//  being accessed by two child threads and a parent thread. //
//                                                           //
//  - Thread th1 enqueues integers, announcing its actions   //
//    as they occur.                                         //
//  - Thread th2 dequeues the integer messages and announces //
//    the result.                                            //
//  - Note that the shared queue is not declared globally,   //
//    It is declared as a data member of mtqueue class.      //
//  - The enQ(val) and deQ() calls are synchronized with the //
//    unique mutex owned by each object of the mtqueue class.//
//    > This has the advantage that, if more than one queue  //
//      is needed, each queue can be synchronized seperately,//
//      so multiple queues can run efficiently.              //
//    > It has advantage that each invocation need not be    //
//      protected since that's done in the member function   //
//      definition.                                          //
//                                                           //
///////////////////////////////////////////////////////////////
//  Build Process                                            //
///////////////////////////////////////////////////////////////
//  Files Required:                                          //
//    queue.cpp, mtqueue.h, mtqueue.cpp                      //
//                                                           //
//  Compiler Command:                                        //
//    cl /GX /MT /D "_X86_" queue.cpp mtqueue.cpp            //
//                                                           //
//  Integrated Environment:                                  //
//    Project\Settings\C++\CodeGeneration\DebugMultithreaded //
//                                                           //
///////////////////////////////////////////////////////////////

#include <windows.h>  // 
#include <process.h>  // _beginthread(...), CriticalSection func's
#include <iostream>   // cout, <<
#include "mtqueue.h"

using namespace std;

CRITICAL_SECTION IOcs;

#define MaxCount 50

//

//----< enqueuing thread function >----------------------------

void __cdecl th1Proc(void* pQueue) {

  int count = 0;
  do {
    int msg = count;

    EnterCriticalSection(&IOcs);
    cout << "\n  sending msg = " << msg;
    LeaveCriticalSection(&IOcs);

    ((mtqueue<int>*)pQueue)->enQ(msg);

    //  You can change relative timing between
    //  threads by using these sleep functions.
    //
        Sleep(1);

  } while(++count < MaxCount);

  EnterCriticalSection(&IOcs);
  cout << "\n\nfinished enQing on thread #1\n";
  LeaveCriticalSection(&IOcs);
}

//----< dequeuing thread function >----------------------------

void __cdecl th2Proc(void *pQueue) {

  mtqueue<int>* pQ = (mtqueue<int>*)pQueue;
  int count = 0;
  while(count++ < MaxCount) {
    while(pQ->length() == 0)
      Sleep(20);

    int msg = pQ->deQ();

    EnterCriticalSection(&IOcs);
    cout << "\n  recving msg = " << msg;
    LeaveCriticalSection(&IOcs);

    //  You can change relative timing between
    //  threads by using these sleep functions.
    //
        Sleep(20);
  };
  EnterCriticalSection(&IOcs);
  cout << "\n\nfinished deQing on thread #2\n";
  LeaveCriticalSection(&IOcs);
}

//

//----< create threads and wait for them to finish >-----------

void main() {

  InitializeCriticalSection(&IOcs);   // protect cout

  mtqueue<int> Q1;

  HANDLE thH[2];
  thH[0] = (HANDLE)_beginthread(th1Proc,0,(void*)&Q1);
  thH[1] = (HANDLE)_beginthread(th2Proc,0,(void*)&Q1);

  EnterCriticalSection(&IOcs);
  cout << "\nwaiting now\n";
  LeaveCriticalSection(&IOcs);

  WaitForMultipleObjects(2,thH,TRUE,INFINITE);
  
  EnterCriticalSection(&IOcs);
  cout << "\n\nending now\n";
  LeaveCriticalSection(&IOcs);

  DeleteCriticalSection(&IOcs);

  cout << endl;
}